@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.17_01';
+our $VERSION = '0.17_02';
 
 use AnyEvent::RipeRedis;
 use AnyEvent::RipeRedis::Error;
@@ -15,7 +15,7 @@ use List::MoreUtils qw( bsearch );
 use Scalar::Util qw( looks_like_number weaken );
 use Carp qw( croak );
 
-our %ERROR_CODES;
+my %ERROR_CODES;
 
 BEGIN {
   %ERROR_CODES = %AnyEvent::RipeRedis::Error::ERROR_CODES;
@@ -75,13 +75,8 @@ my %PREDEFINED_CMDS = (
   sort        => { readonly => 0, key_pos => 1 },
   zunionstore => { readonly => 0, key_pos => 1 },
   zinterstore => { readonly => 0, key_pos => 1 },
-  eval        => { readonly => 0, key_pos => 3 },
-  evalsha     => { readonly => 0, key_pos => 3 },
-  multi       => { readonly => 0, key_pos => 1, force_slot => 1 },
-  watch       => { readonly => 0, key_pos => 1, force_slot => 1 },
-  exec        => { readonly => 0, key_pos => 0, unforce_slot => 1 },
-  discard     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
-  unwatch     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
+  eval        => { readonly => 0, movablekeys => 1, key_pos => 0 },
+  evalsha     => { readonly => 0, movablekeys => 1, key_pos => 0 },
 );
 
 my %SUB_CMDS = (
@@ -131,7 +126,7 @@ sub new {
   $self->{_temp_queue}  = [];
 
   unless ( $self->{lazy} ) {
-    $self->_discover_cluster;
+    $self->_init;
   }
 
   return $self;
@@ -160,8 +155,8 @@ sub disconnect {
 }
 
 sub nodes {
-  my $self = shift;
-  my $key  = shift;
+  my $self         = shift;
+  my $key          = shift;
   my $allow_slaves = shift;
 
   return unless defined $self->{_nodes_pool};
@@ -173,7 +168,9 @@ sub nodes {
 
   my $nodes = $self->_nodes( $slot, $allow_slaves );
 
-  return @{ $self->{_nodes_pool} }{ @{$nodes} };
+  return wantarray
+      ? @{ $self->{_nodes_pool} }{ @{$nodes} }
+      : $self->{_nodes_pool}{ $nodes->[0] };
 }
 
 sub refresh_interval {
@@ -235,21 +232,21 @@ sub crc16 {
 sub hash_slot {
   my $key = shift;
 
-  my $tag = $key;
+  my $hashtag = $key;
 
   if ( $key =~ m/\{([^}]*?)\}/ ) {
     if ( length $1 > 0 ) {
-      $tag = $1;
+      $hashtag = $1;
     }
   }
 
-  return crc16($tag) % MAX_SLOTS;
+  return crc16($hashtag) % MAX_SLOTS;
 }
 
-sub _discover_cluster {
-  my $self  = shift;
+sub _init {
+  my $self = shift;
 
-  $self->{_discover_state} = S_IN_PROGRESS;
+  $self->{_init_state} = S_IN_PROGRESS;
   undef $self->{_refresh_timer};
 
   unless ( defined $self->{_nodes_pool} ) {
@@ -270,12 +267,12 @@ sub _discover_cluster {
 
   weaken($self);
 
-  $self->_cluster_info(
+  $self->_discover_cluster(
     sub {
       my $err = $_[1];
 
       if ( defined $err ) {
-        $self->{_discover_state} = S_NEED_DO;
+        $self->{_init_state} = S_NEED_DO;
 
         $self->{_ready} = 0;
         $self->_abort($err);
@@ -283,7 +280,7 @@ sub _discover_cluster {
         return;
       }
 
-      $self->{_discover_state} = S_DONE;
+      $self->{_init_state} = S_DONE;
 
       $self->{_ready} = 1;
       $self->_process_input_queue;
@@ -292,7 +289,8 @@ sub _discover_cluster {
         $self->{_refresh_timer} = AE::timer(
           $self->{refresh_interval}, 0,
           sub {
-            $self->{_discover_state} = S_NEED_DO;
+            $self->{_init_state} = S_NEED_DO;
+            $self->{_ready}      = 0;
           }
         );
       }
@@ -302,69 +300,50 @@ sub _discover_cluster {
   return;
 }
 
-sub _cluster_info {
+sub _discover_cluster {
   my $self = shift;
   my $cb   = shift;
 
   weaken($self);
 
   $self->_execute(
-    { name => 'cluster_info',
+    { name => 'cluster_state',
       args => [],
 
       on_reply => sub {
-        my $info = shift;
-        my $err  = shift;
+        my $err = $_[1];
 
         if ( defined $err ) {
           $cb->( undef, $err );
           return;
         }
 
-        if ( $info->{cluster_state} eq 'fail' ) {
-          $err = _new_error( 'The cluster is down', E_CLUSTER_DOWN );
-          $cb->( undef, $err );
+        $self->_execute(
+          { name => 'cluster_slots',
+            args => [],
 
-          return;
-        }
+            on_reply => sub {
+              my $slots = shift;
+              my $err   = shift;
 
-        $self->_cluster_slots($cb);
-      }
-    },
-    $self->{_nodes}
-  );
+              if ( defined $err ) {
+                $cb->( undef, $err );
+                return;
+              }
 
-  return;
-};
+              $self->_prepare_nodes( $slots,
+                sub {
+                  unless ( defined $self->{_commands} ) {
+                    $self->_load_commands($cb);
+                    return;
+                  }
 
-sub _cluster_slots {
-  my $self = shift;
-  my $cb   = shift;
-
-  weaken($self);
-
-  $self->_execute(
-    { name => 'cluster_slots',
-      args => [],
-
-      on_reply => sub {
-        my $slots = shift;
-        my $err   = shift;
-
-        if ( defined $err ) {
-          $cb->( undef, $err );
-          return;
-        }
-
-        $self->_prepare_nodes_pool( $slots,
-          sub {
-            unless ( defined $self->{_commands} ) {
-              $self->_load_commands($cb);
-              return;
+                  $cb->();
+                }
+              );
             }
-
-            $cb->();
-          }
+          },
+          $self->{_nodes}
         );
       }
     },
@@ -374,7 +353,7 @@ sub _cluster_slots {
   return;
 }
 
-sub _prepare_nodes_pool {
+sub _prepare_nodes {
   my $self      = shift;
   my $slots_raw = shift;
   my $cb        = shift;
@@ -630,59 +609,38 @@ sub _route {
   my $self = shift;
   my $cmd  = shift;
 
-  if ( $cmd->{name} eq 'multi'
-    && !defined $cmd->{args}[0] )
-  {
-    croak 'Hash tag for "multi" command not specified';
-  }
-
-  if ( $self->{_discover_state} == S_NEED_DO ) {
-    $self->_discover_cluster;
-  }
-
   unless ( $self->{_ready} ) {
+    if ( $self->{_init_state} == S_NEED_DO ) {
+      $self->_init;
+    }
     push( @{ $self->{_input_queue} }, $cmd );
+
     return;
   }
 
+  my $key;
   my $cmd_info = $self->{_commands}{ $cmd->{kwds}[0] };
 
-  my $slot;
-  my $allow_slaves;
+  if ( defined $cmd_info ) {
+    my @tokens = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
 
-  if ( defined $self->{_forced_slot} ) {
-    $slot = $self->{_forced_slot};
-    $cmd->{is_forced_slot} = 1;
-
-    if ( defined $cmd_info && $cmd_info->{unforce_slot} ) {
-      undef $self->{_forced_slot};
-    }
-  }
-  else {
-    my $key;
-
-    if ( defined $cmd_info ) {
-      my @tokens = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
+    if ( $cmd_info->{key_pos} > 0 ) {
       $key = $tokens[ $cmd_info->{key_pos} ];
     }
-
-    $allow_slaves = $self->{allow_slaves};
-
-    if ( defined $key ) {
-      $slot = hash_slot($key);
-
-      if ( $cmd_info->{force_slot}
-        && !defined $self->{_forced_slot} )
-      {
-        $self->{_forced_slot} = $slot;
-      }
-
-      $allow_slaves &&= $cmd_info->{readonly};
+    # Exception for EVAL and EVALSHA commands
+    elsif ( $cmd_info->{movablekeys}
+      && $tokens[2] > 0 )
+    {
+      $key = $tokens[3];
     }
   }
 
-  if ( $cmd->{name} eq 'multi' ) {
-    $cmd->{args} = [];
+  my $slot;
+  my $allow_slaves = $self->{allow_slaves};
+
+  if ( defined $key ) {
+    $slot = hash_slot($key);
+    $allow_slaves &&= $cmd_info->{readonly};
   }
 
   my $nodes = $self->_nodes( $slot, $allow_slaves );
@@ -707,22 +665,37 @@ sub _execute {
   my $hostport = $nodes->[$node_index];
   my $node     = $self->{_nodes_pool}{$hostport};
 
+  my $cmd_name = $cmd->{name} eq 'cluster_state'
+      ? 'cluster_info'
+      : $cmd->{name};
+
   weaken($self);
 
-  $node->execute( $cmd->{name}, @{ $cmd->{args} },
+  $node->execute( $cmd_name, @{ $cmd->{args} },
     { on_reply => sub {
         my $reply = shift;
         my $err   = shift;
 
+        if ( $cmd->{name} eq 'cluster_state' ) {
+          unless ( defined $err ) {
+            if ( $reply->{cluster_state} eq 'ok' ) {
+              $reply = 1;
+            }
+            else {
+              $err = _new_error( 'CLUSTERDOWN The cluster is down',
+                  E_CLUSTER_DOWN );
+            }
+          }
+        }
+
         if ( defined $err ) {
           my $err_code   = $err->code;
           my $nodes_pool = $self->{_nodes_pool};
 
-          if ( ( $err_code == E_MOVED || $err_code == E_ASK )
-            && !$cmd->{is_forced_slot} )
-          {
+          if ( $err_code == E_MOVED || $err_code == E_ASK ) {
             if ( $err_code == E_MOVED ) {
-              $self->_discover_cluster;
+              $self->{_init_state} = S_NEED_DO;
+              $self->{_ready}      = 0;
             }
 
             my ($fwd_hostport) = ( split( m/\s+/, $err->message ) )[2];
@@ -804,16 +777,15 @@ sub _process_input_queue {
 sub _reset_internals {
   my $self = shift;
 
-  $self->{_nodes_pool}     = undef;
-  $self->{_startup_nodes}  = undef;
-  $self->{_nodes}          = undef;
-  $self->{_masters}        = undef;
-  $self->{_slots}          = undef;
-  $self->{_commands}       = undef;
-  $self->{_discover_state} = S_NEED_DO;
-  $self->{_refresh_timer}  = undef;
-  $self->{_ready}          = 0;
-  $self->{_forced_slot}    = undef;
+  $self->{_nodes_pool}    = undef;
+  $self->{_startup_nodes} = undef;
+  $self->{_nodes}         = undef;
+  $self->{_masters}       = undef;
+  $self->{_slots}         = undef;
+  $self->{_commands}      = undef;
+  $self->{_init_state}    = S_NEED_DO;
+  $self->{_refresh_timer} = undef;
+  $self->{_ready}         = 0;
 
   return;
 }
@@ -981,6 +953,7 @@ L<http://redis.io/topics/cluster-spec>
     startup_nodes => [
       { host => 'localhost', port => 7000 },
       { host => 'localhost', port => 7001 },
+      { host => 'localhost', port => 7002 },
     ],
     connection_timeout => 5,
     read_timeout       => 5,
@@ -1268,15 +1241,16 @@ convenient.
 
 =head1 TRANSACTIONS
 
-The client sends all commands of the transaction to the one master node,
-which will be selected by the hash tag in C<MULTI> command or by the first key
-in the C<WATCH> command. The hash tag of the C<MULTI> command is an extension
-of the client and it is not transferred to the Redis server.
+To perform the transaction you must get the master node by the key using
+C<nodes> method and then execute all commands on this node. Nodes must be
+discovered first.
 
-  $cluster->multi('foo');
-  $cluster->set( '{foo}bar', "some\r\nstring" );
-  $cluster->set( '{foo}car', 42 );
-  $cluster->exec(
+  $node = $cluster->nodes('foo');
+
+  $node->multi;
+  $node->set( '{foo}bar', "some\r\nstring" );
+  $node->set( '{foo}car', 42 );
+  $node->exec(
     sub {
       my $reply = shift;
       my $err   = shift;
@@ -1322,7 +1296,8 @@ aborted.
 
 =head2 nodes( [ $key ] [, $allow_slaves ] )
 
-Gets particular nodes of the cluster. Nodes must be discovered first.
+Gets particular nodes of the cluster. Nodes must be discovered first. In scalar
+context method returns the first node from the list.
 
 Getting all master nodes of the cluster:
 
@@ -1332,9 +1307,9 @@ Getting all nodes of the cluster, including slave nodes:
 
   my @nodes = $cluster->nodes( undef, 1 );
 
-Getting master nodes by the key:
+Getting master node by the key:
 
-  my @master_nodes = $cluster->nodes('foo');
+  my $master_node = $cluster->nodes('foo');
 
 Getting nodes by the key, including slave nodes:
 
